"""Image drawing primitives"""
from typing import Any, Tuple

from PIL import ImageDraw  # type: ignore


def gen_large_sun(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Generate large sun icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    line_width: int = 5
    # Protruding rays
    draw.polygon(
        (
            (x_y[0] + 29, x_y[1]),
            (x_y[0] + 34, x_y[1] + 16),
            (x_y[0] + 24, x_y[1] + 16),
            (x_y[0] + 29, x_y[1]),
        ),
        color,
        line_width,
    )  # Top
    draw.polygon(
        (
            (x_y[0] + 29, x_y[1] + 56),
            (x_y[0] + 34, x_y[1] + 40),
            (x_y[0] + 24, x_y[1] + 40),
            (x_y[0] + 29, x_y[1] + 56),
        ),
        color,
        line_width,
    )  # Bottom
    draw.polygon(
        (
            (x_y[0], x_y[1] + 28),
            (x_y[0] + 17, x_y[1] + 23),
            (x_y[0] + 17, x_y[1] + 33),
            (x_y[0], x_y[1] + 28),
        ),
        color,
        line_width,
    )  # Left
    draw.polygon(
        (
            (x_y[0] + 57, x_y[1] + 28),
            (x_y[0] + 40, x_y[1] + 23),
            (x_y[0] + 40, x_y[1] + 33),
            (x_y[0] + 57, x_y[1] + 28),
        ),
        color,
        line_width,
    )  # Right
    draw.line(
        (x_y[0] + 10, x_y[1] + 10, x_y[0] + 47, x_y[1] + 47),
        color,
        line_width,
    )
    draw.line(
        (x_y[0] + 10, x_y[1] + 47, x_y[0] + 47, x_y[1] + 10),
        color,
        line_width,
    )
    # Sun circle
    draw.ellipse((x_y[0] + 12, x_y[1] + 12, x_y[0] + 45, x_y[1] + 45), color_neg)
    draw.ellipse((x_y[0] + 17, x_y[1] + 17, x_y[0] + 40, x_y[1] + 40), color)


def gen_small_sun(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Generate small sun icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    line_width: int = 5
    # Protruding rays
    draw.line((x_y[0] + 5, x_y[1] + 5, x_y[0] + 25, x_y[1] + 25), color, line_width)
    draw.line((x_y[0] + 5, x_y[1] + 25, x_y[0] + 25, x_y[1] + 5), color, line_width)
    draw.polygon(
        (
            (x_y[0] + 15, x_y[1]),
            (x_y[0] + 10, x_y[1] + 10),
            (x_y[0] + 20, x_y[1] + 10),
        ),
        color,
        line_width,
    )  # top
    draw.polygon(
        (
            (x_y[0] + 15, x_y[1] + 30),
            (x_y[0] + 10, x_y[1] + 20),
            (x_y[0] + 20, x_y[1] + 20),
        ),
        color,
        line_width,
    )  # bottom
    draw.polygon(
        (
            (x_y[0] + 10, x_y[1] + 10),
            (x_y[0], x_y[1] + 15),
            (x_y[0] + 20, x_y[1] + 20),
        ),
        color,
        line_width,
    )  # left
    draw.polygon(
        (
            (x_y[0] + 20, x_y[1] + 10),
            (x_y[0] + 30, x_y[1] + 15),
            (x_y[0] + 20, x_y[1] + 20),
        ),
        color,
        line_width,
    )  # right
    # Sun circle
    draw.ellipse((x_y[0] + 5, x_y[1] + 5, x_y[0] + 25, x_y[1] + 25), color)
    draw.ellipse((x_y[0] + 10, x_y[1] + 10, x_y[0] + 20, x_y[1] + 20), color_neg)


def gen_large_cloud(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Generate large cloud icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    # Outline
    draw.ellipse((x_y[0], x_y[1] + 20, x_y[0] + 20, x_y[1] + 40), color)
    draw.ellipse((x_y[0] + 5, x_y[1] + 10, x_y[0] + 35, x_y[1] + 40), color)
    draw.ellipse((x_y[0] + 15, x_y[1], x_y[0] + 55, x_y[1] + 40), color)
    draw.ellipse((x_y[0] + 35, x_y[1] + 10, x_y[0] + 65, x_y[1] + 40), color)
    # Negative Space
    draw.ellipse((x_y[0] + 5, x_y[1] + 25, x_y[0] + 15, x_y[1] + 35), color_neg)
    draw.ellipse((x_y[0] + 10, x_y[1] + 15, x_y[0] + 30, x_y[1] + 35), color_neg)
    draw.ellipse((x_y[0] + 20, x_y[1] + 5, x_y[0] + 50, x_y[1] + 35), color_neg)
    draw.ellipse((x_y[0] + 40, x_y[1] + 15, x_y[0] + 60, x_y[1] + 35), color_neg)


def gen_small_cloud(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Generate small cloud icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    # Outline
    draw.ellipse((x_y[0], x_y[1] + 10, x_y[0] + 11, x_y[1] + 21), color)
    draw.ellipse((x_y[0] + 5, x_y[1] + 5, x_y[0] + 21, x_y[1] + 21), color)
    draw.ellipse((x_y[0] + 10, x_y[1], x_y[0] + 31, x_y[1] + 21), color)
    draw.ellipse((x_y[0] + 20, x_y[1] + 5, x_y[0] + 36, x_y[1] + 21), color)
    # Negative Space
    draw.ellipse((x_y[0] + 3, x_y[1] + 13, x_y[0] + 8, x_y[1] + 18), color_neg)
    draw.ellipse((x_y[0] + 8, x_y[1] + 8, x_y[0] + 18, x_y[1] + 18), color_neg)
    draw.ellipse((x_y[0] + 13, x_y[1] + 3, x_y[0] + 28, x_y[1] + 18), color_neg)
    draw.ellipse((x_y[0] + 23, x_y[1] + 8, x_y[0] + 33, x_y[1] + 18), color_neg)


def gen_raindrop(draw: ImageDraw, color: Any, x_y: Tuple[int, int]) -> None:
    """Generate raindrop icon

    Args:
        draw: ImageDraw object
        color: Color of the object
        x_y: (x, y) coordinates
    """
    # Tail
    draw.ellipse((x_y[0] + 3, x_y[1] + 3, x_y[0] + 8, x_y[1] + 7), color)
    # Head
    draw.polygon(
        (
            (x_y[0], x_y[1]),
            (x_y[0] + 6, x_y[1] + 6),
            (x_y[0] + 7, x_y[1] + 3),
            (x_y[0] + 3, x_y[1]),
        ),
        color,
    )


def gen_lightning(draw: ImageDraw, color: Any, x_y: Tuple[int, int]) -> None:
    """Generate lightning bolt icon

    Args:
        draw: ImageDraw object
        color: Color of the object
        x_y: (x, y) coordinates
    """
    draw.polygon(
        (
            (x_y[0], x_y[1]),
            (x_y[0] + 8, x_y[1]),
            (x_y[0] + 12, x_y[1] + 6),
            (x_y[0] + 6, x_y[1] + 6),
            (x_y[0] + 8, x_y[1] + 12),
            (x_y[0], x_y[1] + 4),
            (x_y[0] + 7, x_y[1] + 4),
            (x_y[0], x_y[1]),
        ),
        color,
    )


def gen_snowflake(draw: ImageDraw, color: Any, x_y: Tuple[int, int]) -> None:
    """Generate snowflake icon

    Args:
        draw: ImageDraw object
        color: Color of the object
        x_y: (x, y) coordinates
    """
    line_width: int = 2
    draw.line((x_y[0] + 5, x_y[1], x_y[0] + 5, x_y[1] + 8), color, line_width)
    draw.line((x_y[0] + 1, x_y[1] + 1, x_y[0] + 10, x_y[1] + 6), color, line_width)
    draw.line((x_y[0] + 1, x_y[1] + 6, x_y[0] + 10, x_y[1] + 1), color, line_width)


def gen_mist(draw: ImageDraw, color: Any, x_y: Tuple[int, int]) -> None:
    """Generate mist icon

    Args:
        draw: ImageDraw object
        color: Color of the object
        x_y: (x, y) coordinates
    """
    line_width: int = 4
    draw.line((x_y[0] + 22, x_y[1], x_y[0] + 40, x_y[1]), color, line_width)
    draw.line((x_y[0] + 4, x_y[1] + 8, x_y[0] + 47, x_y[1] + 8), color, line_width)
    draw.line(
        (x_y[0] + 15, x_y[1] + 16, x_y[0] + 60, x_y[1] + 16),
        color,
        line_width,
    )
    draw.line((x_y[0], x_y[1] + 24, x_y[0] + 55, x_y[1] + 24), color, line_width)
    draw.line((x_y[0] + 9, x_y[1] + 32, x_y[0] + 51, x_y[1] + 32), color, line_width)
    draw.line(
        (x_y[0] + 20, x_y[1] + 40, x_y[0] + 40, x_y[1] + 40),
        color,
        line_width,
    )


def gen_closed_eye_icon(draw: ImageDraw, color: Any, x_y: Tuple[int, int]):
    """Generate closed eye icon at center of x_y

    Args:
        draw: ImageDraw object
        color: Color of the object
        x_y: (x, y) coordinates of center point
    """
    line_width: int = 8
    x_0, y_0 = x_y[0] - 75, -50
    x_1, y_1 = x_y[0] + 75, x_y[1] - 75
    a_start, a_end = 20, 160
    draw.arc([(x_0, y_0), (x_1, y_1)], a_start, a_end, color, line_width)
    draw.line([(x_0 + 9, y_0 + 131), (x_0 + 29, y_0 + 111)], color, line_width)
    draw.line([(x_0 + 49, y_0 + 147), (x_0 + 59, y_0 + 122)], color, line_width)
    draw.line([(x_0 + 104, y_0 + 147), (x_0 + 94, y_0 + 122)], color, line_width)
    draw.line([(x_0 + 144, y_0 + 131), (x_0 + 124, y_0 + 111)], color, line_width)



"""
Test double fake classes
"""
from typing import Optional


class FakeResponse:
    """Fake response object for use in returning fake json data

    See FakeRequests class for more info
    """

    def __init__(self, json_data: dict, status_code: int) -> None:
        """Initialize fake response object

        Args:
            json_data (dict): json data to return
            status_code (str): status code to return
        """
        self.json_data: dict = json_data
        self.status_code: int = status_code

    def json(self) -> dict:
        """Return json data as a dictionary"""
        return self.json_data


class FakeRequests:
    """Fake requests object. Used to mock requests.get() in tests.

    Usage:
    >>> import json
    >>> file_name = "your_file.json"
    >>> requests = FakeRequests()
    >>> with open(file_name, "r", encoding="utf-8") as file:
    >>>     test_data = json.load(file)
    >>>     requests.add_response(test_data, 200)
    >>> response = requests.get("https://www.example.com")
    >>> assert response ...
    """

    def __init__(self) -> None:
        """Initialize fake requests object"""
        self.response: Optional[FakeResponse] = None

    def add_response(self, json_data: dict, status_code: int) -> None:
        """Set fake response

        Args:
            json_data (dict): json data
            status_code (str): status code
        """
        self.response = FakeResponse(json_data, status_code)

    # pylint: disable=unused-argument
    def get(self, url: str, params: dict = None) -> Optional[FakeResponse]:
        """Fake get method

        Args:
            url (str): url
            params (dict): params
        """
        assert url is not None
        assert self.response is not None
        return self.response



"""Base class and helper functions for train model"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, Optional

from loguru import logger


def abbreviate_stn_name(station_name: str) -> str:
    """Helper function to abbreviate station name by shortening words

    Args:
        station_name (str): Station name

    Returns:
        str: Abbreviated station name
    """
    abbreviation_dict: Dict[str, str] = {
        "Station": "Stn",
        "Street": "St",
        "Lane": "Ln",
        "Court": "Ct",
        "Road": "Rd",
        "North": "N",
        "South": "S",
        "East": "E",
        "West": "W",
        "Thameslink": "TL",
    }
    for word, abbreviation in abbreviation_dict.items():
        station_name = station_name.replace(word, abbreviation)

    return station_name


class TrainModel(Enum):
    """Enum of train models"""

    HUXLEY2 = auto()
    OPEN_LIVE = auto()


@dataclass
class TrainObject:
    """Train object"""

    model: TrainModel
    station_from: str
    station_to: str
    number: int
    url: str = ""
    token: str = ""


class TrainBase(ABC):
    """Abstract base class for all train models"""

    def __init__(self):
        self._num: int = 0
        self._data: Optional[Any] = None
        self.origin: str = ""
        self.destination: str = ""

    @abstractmethod
    def retrieve_data(self, protocol: Any, train_object: TrainObject) -> None:
        """Retrieves train data from API; must be called after constructor

        Args:
            protocol: HTTP data protocol (e.g. requests or zeep)
            train_object: Train object
        """

    @abstractmethod
    def fetch_train(self, num: int) -> str:
        """Return requested train data

        Args:
            num (int): Desired train number

        Returns:
            str: Train data
        """

    @staticmethod
    def format_train_string(
        arrival_t: str, platform: str, dest_stn: str, status: str
    ) -> str:
        """Takes in a train's arrival time, platform, destination station, and status,
        and returns a string with the train's information in a display format.

        Args:
            arrival_t (str): The time the train is due to arrive at the station.
            platform (str): The platform number the train is arriving at.
            dest_stn (str): The destination station of the train.
            status (str): The status of the train.

        Returns:
            str: Formatted string
        """
        return (
            f"{arrival_t} | P{platform} to {abbreviate_stn_name(dest_stn)} - {status}"
        )

    @staticmethod
    def format_error_msg(error_msg: str, num: int) -> str:
        """Format error message by line wrapping over each line

        Args:
            error_msg (str): Error message
            num (int): Train number

        Returns:
            str: Formatted error message if num is 1 else empty string
        """
        line_length: int = 38
        if num == 0:
            logger.error(error_msg)
        return (error_msg[num * line_length : (num + 1) * line_length]).lstrip(" ")

    def _validate_number(self, num: int) -> None:
        """Check if train number is valid

        Args:
            num (int): Train number to check

        Raises:
            ValueError: Invalid train number
        """
        if num < 0 or num > self._num:
            raise ValueError(
                f"{num} is an invalid train request number (max: {self._num})"
            )



"""Base class and helper functions for weather model"""
# pylint: disable=duplicate-code
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any


def kelvin_to_celsius(kelvin_temp: float) -> float:
    """Helper function to convert Kelvin to Celsius to one decimal place"""
    if kelvin_temp < 0:
        raise ValueError("Kelvin temperature cannot be less than 0.")
    return round(kelvin_temp - 273.15, 1)


def celsius_to_fahrenheit(celsius_temp: float) -> float:
    """Helper function to convert Celsius to Fahrenheit to one decimal place"""
    if celsius_temp < -273.15:
        raise ValueError("Celsius temperature cannot be less than -273.15.")
    return round((celsius_temp * 9 / 5) + 32.0, 1)


class WeatherModel(Enum):
    """Enum of weather models"""

    OPEN_WEATHER_MAP = auto()


@dataclass
class WeatherObject:
    """Weather object"""

    model: WeatherModel
    latitude: float
    longitude: float
    exclude_flags: str
    weather_api_token: str


class IconType(Enum):
    """Enum for Weather Icon types"""

    CLEAR_SKY = auto()
    FEW_CLOUDS = auto()
    SCATTERED_CLOUDS = auto()
    BROKEN_CLOUDS = auto()
    SHOWER_RAIN = auto()
    RAIN = auto()
    THUNDERSTORM = auto()
    SNOW = auto()
    MIST = auto()


class ScaleType(Enum):
    """Enum for Weather Scale types"""

    CELSIUS = auto()
    FAHRENHEIT = auto()


class WeatherBase(ABC):
    """Abstract base class for all weather models"""

    @abstractmethod
    def retrieve_data(self, protocol: Any, weather_object: WeatherObject) -> None:
        """Retrieves weather data from API; must be called after constructor

        Args:
            protocol: HTTP data protocol (e.g. requests)
            weather_object: Weather object
        """

    @abstractmethod
    def get_icon(self, day: int = 0) -> IconType:
        """Return requested weather icon

        Args:
            day (int): Desired day number (0/today or 1..7)

        Returns:
            IconType: Weather IconType
        """

    @abstractmethod
    def get_current_weather(self, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Return requested weather data

        Args:
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """

    @abstractmethod
    def get_current_temperature(self, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Return requested current temperature

        Args:
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """

    @abstractmethod
    def get_current_condition(self) -> str:
        """Return requested current condition

        Returns:
            str: Formatted string or error message
        """

    @abstractmethod
    def get_temp_range(self, day: int, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Return temperature range string

        Args:
            day (int): Desired day number (0/today or 1..7)
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """

    @abstractmethod
    def get_condition(self, day: int) -> str:
        """Return weather condition string

        Args:
            day (int): Desired day number (0/today or 1/tomorrow)

        Returns:
            str: Formatted string or error message
        """

    @abstractmethod
    def get_future_weather(self, day: int, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Return weather string for given day

        Args:
            day (int): Desired day number (0/today or 1..7)
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """



"""Weather icon drawing functions"""
from typing import Any, Tuple

from PIL import ImageDraw  # type: ignore

from inky_pi.display.util.shapes import (
    gen_large_cloud,
    gen_large_sun,
    gen_lightning,
    gen_mist,
    gen_raindrop,
    gen_small_cloud,
    gen_small_sun,
    gen_snowflake,
)


def draw_sun_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large sun icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_sun(draw, color, color_neg, (x_y[0] + 7, x_y[1]))


def draw_sun_cloud_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw small sun + large cloud icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_small_sun(draw, color, color_neg, x_y)
    gen_large_cloud(draw, color, color_neg, (x_y[0], x_y[1] + 5))


def draw_cloud_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large cloud

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_cloud(draw, color, color_neg, x_y)


def draw_two_clouds_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large cloud + small cloud icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_cloud(draw, color, color_neg, x_y)
    gen_small_cloud(draw, color, color_neg, (x_y[0] + 30, x_y[1] + 25))


def draw_cloud_rain_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large cloud + two rain drop icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_cloud(draw, color, color_neg, x_y)
    gen_raindrop(draw, color, (x_y[0] + 25, x_y[1] + 45))
    gen_raindrop(draw, color, (x_y[0] + 42, x_y[1] + 45))


def draw_sun_cloud_rain_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw small sun, large cloud + two rain drop icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_small_sun(draw, color, color_neg, x_y)
    gen_large_cloud(draw, color, color_neg, (x_y[0], x_y[1] + 5))
    gen_raindrop(draw, color, (x_y[0] + 25, x_y[1] + 50))
    gen_raindrop(draw, color, (x_y[0] + 42, x_y[1] + 50))


def draw_cloud_lightning_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large cloud + lightning bolt icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_cloud(draw, color, color_neg, x_y)
    gen_lightning(draw, color, (x_y[0] + 30, x_y[1] + 45))


def draw_cloud_snow_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw large cloud + snowflake icons

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_large_cloud(draw, color, color_neg, x_y)
    gen_snowflake(draw, color, (x_y[0] + 12, x_y[1] + 45))
    gen_snowflake(draw, color, (x_y[0] + 26, x_y[1] + 50))
    gen_snowflake(draw, color, (x_y[0] + 40, x_y[1] + 45))


# pylint: disable=unused-argument
def draw_mist_icon(
    draw: ImageDraw, color: Any, color_neg: Any, x_y: Tuple[int, int]
) -> None:
    """Draw mist icon

    Args:
        draw: ImageDraw object
        color: Color of the outlines
        color_neg: Color of the negative space
        x_y: (x, y) coordinates
    """
    gen_mist(draw, color, (x_y[0] + 5, x_y[1]))



"""
Performs end-to-end test with terminal output to check full program functionality.
Actively retrieves data from both weather and train APIs for testing. Tests may fail
if API endpoints are down.
"""
import re
from typing import List
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner, Result

from inky_pi.cli import cli
from inky_pi.display.util.desktop_driver import DesktopDisplayDriver


@pytest.mark.e2e
def test_running_terminal_ui_generates_expected_output() -> None:
    """Runs the full program and checks each test string is present in output"""
    time = r"(\d+(?:\:\d+))"
    temp = r"(\d+(?:\.\d+)?°(C|F))"
    station_abbr = r"\b[A-Z]{3}(?![A-Z])"
    error_type1 = r"There\sare\sno\strain\sservices\sat\sthis\s\w+"
    error_type2 = r"No\strains\sto\s\w+"
    any_error_msg = rf"({error_type1})|({error_type2})"

    test_regex_list: List[str] = [
        # Time (HH:MM)
        rf"{time}",
        # Date (Day dd MMM yyyy)
        r"[A-Z][a-z]{2}\s\d*\s[A-Z][a-z]*\s\d{4}",
        # now: <current>°C/F
        rf"now:\s{temp}",
        # today: <low>°C/F – <high>°C/F
        rf"today:\s{temp}\s–\s{temp}",
        # train schedule from <station> to <station>
        rf"train\sschedule\sfrom\s{station_abbr}\sto\s{station_abbr}:",
        # <HH:MM> | P<#> to <station> - <status> [OR] <error message>
        rf"({time}\s\|\sP\d+\sto\s\w+\s\-\s(\w+\s|{time}))|({any_error_msg})",
    ]

    result: Result = CliRunner().invoke(
        cli, ["display", "--option", "train", "--output", "terminal"]
    )
    assert result.exit_code == 0
    for test_regex in test_regex_list:
        assert re.compile(test_regex).search(result.output) is not None


@pytest.mark.e2e
@patch("inky_pi.display.util.desktop_driver.DesktopDisplayDriver.show")
def test_running_desktop_ui_generates_expected_output(image_show_mock: Mock) -> None:
    """Runs the full program and checks image is drawn to desktop window

    Args:
        image_show_mock (Mock): Mock for image show method
    """
    result: Result = CliRunner().invoke(cli, ["display", "--output", "desktop"])
    image_show_mock.assert_called_once()
    assert result.exit_code == 0


@pytest.mark.e2e
@patch("inky_pi.display.util.desktop_driver.DesktopDisplayDriver.show")
@patch("inky_pi.display.inky_draw._import_inky_what")
def test_running_inky_train_ui_generates_expected_output(
    _import_inky_what_mock: Mock,
    image_show_mock: Mock,
) -> None:
    """Runs the full program and checks image is drawn to desktop window

    Args:
        _import_inky_what_mock (Mock): Mock for import inky_what
        image_show_mock (Mock): Mock for image show method
    """
    _import_inky_what_mock.return_value = DesktopDisplayDriver
    result: Result = CliRunner().invoke(
        cli, ["display", "--option", "train", "--output", "desktop"]
    )
    image_show_mock.assert_called_once()
    assert result.exit_code == 0


@pytest.mark.e2e
@patch("inky_pi.display.util.desktop_driver.DesktopDisplayDriver.show")
@patch("inky_pi.display.inky_draw._import_inky_what")
def test_running_inky_weather_ui_generates_expected_output(
    _import_inky_what_mock: Mock,
    image_show_mock: Mock,
) -> None:
    """Runs the full program and checks image is drawn to desktop window

    Args:
        _import_inky_what_mock (Mock): Mock for import inky_what
        image_show_mock (Mock): Mock for image show method
    """
    _import_inky_what_mock.return_value = DesktopDisplayDriver
    result: Result = CliRunner().invoke(
        cli, ["display", "--option", "weather", "--output", "desktop"]
    )
    image_show_mock.assert_called_once()
    assert result.exit_code == 0


@pytest.mark.e2e
@patch("inky_pi.display.util.desktop_driver.DesktopDisplayDriver.show")
@patch("inky_pi.display.inky_draw._import_inky_what")
def test_running_inky_night_ui_generates_expected_output(
    _import_inky_what_mock: Mock,
    image_show_mock: Mock,
) -> None:
    """Runs the full program and checks image is drawn to desktop window

    Args:
        _import_inky_what_mock (Mock): Mock for import inky_what
        image_show_mock (Mock): Mock for image show method
    """
    _import_inky_what_mock.return_value = DesktopDisplayDriver
    result: Result = CliRunner().invoke(
        cli, ["display", "--option", "night", "--output", "desktop"]
    )
    image_show_mock.assert_called_once()
    assert result.exit_code == 0



"""Test cli commands"""
import pytest
from click.testing import CliRunner

from inky_pi.__init__ import __version__  # type: ignore
from inky_pi.cli import OUTPUT_PREFIX, cli


def test_cli_main() -> None:
    """Tests inky_pi main command outputs expected result"""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code == 0
    assert "Usage: cli [OPTIONS] COMMAND [ARGS]" in result.output


@pytest.mark.parametrize(
    "command, expected_output",
    [
        (
            ["display", "--option", "train", "--output", "terminal", "--dry-run"],
            f"{OUTPUT_PREFIX} DisplayOption.TRAIN DisplayModel.TERMINAL",
        ),
        (
            ["display", "--option", "weather", "--output", "Terminal", "--dry-run"],
            f"{OUTPUT_PREFIX} DisplayOption.WEATHER DisplayModel.TERMINAL",
        ),
        (
            ["display", "--option", "night", "--output", "TERMINAL", "--dry-run"],
            f"{OUTPUT_PREFIX} DisplayOption.NIGHT DisplayModel.TERMINAL",
        ),
        (
            ["display", "--option", "Train", "--output", "inky", "--dry-run"],
            f"{OUTPUT_PREFIX} DisplayOption.TRAIN DisplayModel.INKY_WHAT",
        ),
        (
            ["display", "--option", "TRAIN", "--output", "desktop", "--dry-run"],
            f"{OUTPUT_PREFIX} DisplayOption.TRAIN DisplayModel.DESKTOP",
        ),
        (
            ["display", "--version"],
            f"cli, version {__version__}",
        ),
    ],
)
def test_cli_args(command, expected_output) -> None:
    """Tests inky_pi command with various args outputs expected result

    Args:
        command (list): command to run
        expected_output (str): expected output
    """
    runner = CliRunner()
    result = runner.invoke(cli, command)
    assert result.exit_code == 0
    assert result.output == f"{expected_output}\n"



"""Tests for display module"""
import platform
from unittest.mock import Mock, patch

import pytest

from inky_pi.display.display_base import DisplayBase, DisplayModel, DisplayObject
from inky_pi.display.inky_draw import InkyDraw
from inky_pi.display.terminal_draw import TerminalDraw
from inky_pi.util import display_model_factory, import_display


@patch("inky_pi.display.inky_draw._import_inky_what")
@patch("inky_pi.display.inky_draw.InkyDraw")
def test_can_successfully_instantiate_inky_draw_object(
    mock_inky_draw: Mock, mock_inky_what: Mock
) -> None:
    """Test for creating InkyDraw instanced object

    Args:
        mock_inky_draw: Mock for InkyDraw
        mock_inky_what: Mock for _import_inky_what
    """

    inky_object = DisplayObject(model=DisplayModel.INKY_WHAT, base_color="black")
    import_display(inky_object)
    mock_inky_draw.assert_called_once()
    mock_inky_what.assert_called_once()


def test_instantiating_inky_draw_object_not_on_rpi_raises_import_error() -> None:
    """In on a Raspberry Pi, should be able to instantiate an inky display object;
    but if not, should get an ImportError
    """

    inky_object = DisplayObject(model=DisplayModel.INKY_WHAT, base_color="black")
    if platform.machine() == "armv7l":
        ret: DisplayBase = import_display(inky_object)
        assert isinstance(ret, InkyDraw)
    else:
        with pytest.raises(ImportError):
            display_model_factory(inky_object)
        with pytest.raises(SystemExit) as err:
            import_display(inky_object)
            assert err.type == SystemExit
            assert err.value.code == 1


def test_can_successfully_instantiate_terminal_draw_object() -> None:
    """Test for creating terminal instanced object"""

    terminal_object = DisplayObject(model=DisplayModel.TERMINAL)
    ret: DisplayBase = import_display(terminal_object)
    assert isinstance(ret, TerminalDraw)


def test_can_successfully_instantiate_desktop_draw_object() -> None:
    """Test for creating desktop instanced object"""

    desktop_object = DisplayObject(model=DisplayModel.DESKTOP, base_color="yellow")
    ret: DisplayBase = import_display(desktop_object)
    assert isinstance(ret, InkyDraw)



"""Inky_Pi weather model module.

Fetches data from OpenWeatherMap API and generates formatted data"""
from typing import Any, Dict, Union

import requests
from loguru import logger

from inky_pi.weather.weather_base import (
    IconType,
    ScaleType,
    WeatherBase,
    WeatherObject,
    celsius_to_fahrenheit,
    kelvin_to_celsius,
)

# Weather formatting constants
DEG_C: str = "\N{DEGREE SIGN}" + "C"
DEG_F: str = "\N{DEGREE SIGN}" + "F"


def _check_day_limit(day: int) -> None:
    if day < 0 or day >= 8:
        logger.error("Invalid day requested", day)
        raise ValueError(
            "Weather data only available for 0 (today) or up to 7 days ahead."
        )


class OpenWeatherMap(WeatherBase):
    """Fetch and manage weather data"""

    def __init__(self) -> None:
        """Initialize variables"""
        self._data: dict = {}

    def retrieve_data(self, protocol: Any, weather_object: WeatherObject) -> None:
        """Retrieves weather data from OpenWeatherMap 7-day forecast API.
        This must be called before any other data manipulation methods.

        Args:
            protocol (Any): Requests object
            weather_object: WeatherObject object
        """
        payload: Dict[str, Union[float, str]] = {
            "lat": weather_object.latitude,
            "lon": weather_object.longitude,
            "exclude": weather_object.exclude_flags,
            "appid": weather_object.weather_api_token,
        }
        response: Any = protocol.get(
            "https://api.openweathermap.org/data/2.5/onecall?", params=payload
        )

        self._data = response.json()

        # Check for errors in weather response, i.e. API key invalid (cod==401)
        if "cod" in self._data:
            err_message: str = self._data["message"]
            logger.warning("Error in weather response", err_message)
            raise ValueError(err_message)

    def get_icon(self, day: int = 0) -> IconType:
        """Retrieves weather type from current OpenWeatherMap weather icon

        Full list of icons/codes: https://openweathermap.org/weather-conditions

        Args:
            day (int): Desired day number (0/today or 1..7)

        Returns:
            IconType: Weather IconType
        """
        _check_day_limit(day)
        icon_code: str = ""
        # Get first 2 code characters; 3rd character is 'd/n' for day/night (ignored)
        if day == 0:
            icon_code = str(self._data["current"]["weather"][0]["icon"])[0:2]
        else:
            icon_code = str(self._data["daily"][day]["weather"][0]["icon"])[0:2]

        weather_type_dict: Dict[str, IconType] = {
            "01": IconType.CLEAR_SKY,
            "02": IconType.FEW_CLOUDS,
            "03": IconType.SCATTERED_CLOUDS,
            "04": IconType.BROKEN_CLOUDS,
            "09": IconType.SHOWER_RAIN,
            "10": IconType.RAIN,
            "11": IconType.THUNDERSTORM,
            "13": IconType.SNOW,
            "50": IconType.MIST,
        }
        return weather_type_dict[icon_code]

    def get_current_weather(self, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Generate current weather string

        String is returned in format:
            [XX.X]°[C/F] - [Current Weather]

        Args:
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """
        return f"{self.get_current_temperature(scale)} - {self.get_current_condition()}"

    def get_current_temperature(self, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Generate current temperature

        Args:
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted temperature string or error message
        """
        try:
            celsius_temp: float = kelvin_to_celsius(
                float(self._data["current"]["temp"])
            )
            formatted_temp: str = (
                str(celsius_temp) + DEG_C
                if scale == ScaleType.CELSIUS
                else str(celsius_to_fahrenheit(celsius_temp)) + DEG_F
            )
            return f"{formatted_temp}"
        except (KeyError, IndexError) as ex:
            logger.error("Invalid get_current_weather data", repr(ex))
            return f"Error retrieving temperature. {ex!r}"

    def get_current_condition(self) -> str:
        """Generate current weather condition

        Returns:
            str: Condition or error message
        """
        try:
            str_status: str = self._data["current"]["weather"][0]["main"]
            return str_status
        except (KeyError, IndexError) as ex:
            logger.error("Invalid get_current_weather data", repr(ex))
            return f"Error retrieving condition. {ex!r}"

    def get_temp_range(self, day: int, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Generate temperature range string

        String is returned in format:
            [XX.X(min)]°[C/F] – [XX.X(max)]°[C/F]

        Args:
            day (int): Desired day number (0/today or 1/tomorrow)
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """
        _check_day_limit(day)
        try:
            celsius_temp_min: float = kelvin_to_celsius(
                float(self._data["daily"][day]["temp"]["min"])
            )
            celsius_temp_max: float = kelvin_to_celsius(
                float(self._data["daily"][day]["temp"]["max"])
            )
            str_temp_min: str = (
                str(celsius_temp_min) + DEG_C
                if scale == ScaleType.CELSIUS
                else str(celsius_to_fahrenheit(celsius_temp_min)) + DEG_F
            )
            str_temp_max: str = (
                str(celsius_temp_max) + DEG_C
                if scale == ScaleType.CELSIUS
                else str(celsius_to_fahrenheit(celsius_temp_max)) + DEG_F
            )
            return f"{str_temp_min} – {str_temp_max}"
        except (KeyError, IndexError) as ex:
            logger.error("Invalid get_temp_range data", repr(ex))
            return f"Error retrieving range. {ex!r}"

    def get_condition(self, day: int) -> str:
        """Generate weather condition string

        Args:
            day (int): Desired day number (0/today or 1/tomorrow)

        Returns:
            str: Formatted string or error message
        """
        _check_day_limit(day)
        try:
            return f"{self._data['daily'][day]['weather'][0]['description']}"

        except (KeyError, IndexError) as ex:
            logger.error("Invalid fetch_condition data", repr(ex))
            return f"Error retrieving condition. {ex!r}"

    def get_future_weather(self, day: int, scale: ScaleType = ScaleType.CELSIUS) -> str:
        """Generate weather string for given day

        String is returned in format:
            [XX.X]°[C/F]

        Args:
            day (int): Desired day number (0/today or 1..7)
            scale (ScaleType): Celsius or Fahrenheit for formatting

        Returns:
            str: Formatted string or error message
        """
        _check_day_limit(day)
        try:
            celsius_temp: float = kelvin_to_celsius(
                float(self._data["daily"][day]["temp"]["day"])
            )
            str_temp: str = (
                str(celsius_temp) + DEG_C
                if scale == ScaleType.CELSIUS
                else str(celsius_to_fahrenheit(celsius_temp)) + DEG_F
            )
            return f"{str_temp}"
        except (KeyError, IndexError) as ex:
            logger.error("Invalid get_current_weather data", repr(ex))
            return f"Error retrieving weather. {ex!r}"


def instantiate_open_weather_map(weather_object: WeatherObject) -> OpenWeatherMap:
    """Open Weather Map object creator

    Args:
        weather_object (WeatherObject): weather object containing model

    Returns:
        OpenWeatherMap: OpenWeatherMap object
    """
    weather_base = OpenWeatherMap()
    weather_base.retrieve_data(requests, weather_object)
    return weather_base



"""Console script for inky_pi."""
import click
from loguru import logger

from inky_pi.__init__ import __version__  # type: ignore
from inky_pi.__main__ import OUTPUT_HANDLER, DisplayOption, display_data
from inky_pi.util import configure_logging

OUTPUT_PREFIX = "inky_pi cli"


@click.group()
def cli():
    """CLI group for inky_pi."""


@cli.command()
@click.version_option(version=__version__)
@click.option(
    "-o", "--option", default="train", help="Display option (train, weather, night)"
)
@click.option(
    "-m", "--output", default="inky", help="Output source (inky, terminal, desktop)"
)
@click.option("--dry-run", is_flag=True, default=False, help="Dry run")
def display(option: str, output: str, dry_run: bool):
    """Console script for inky_pi train and weather."""
    click.echo(
        f"{OUTPUT_PREFIX}"
        f" {DisplayOption[option.upper()]}"
        f" {OUTPUT_HANDLER[output.lower()].model}"
    )
    if not dry_run:
        display_data(DisplayOption[option.upper()], OUTPUT_HANDLER[output.lower()])


def main():
    """CLI main method."""
    configure_logging()
    logger.debug("InkyPi cli initialized")
    cli()


if __name__ == "__main__":
    main()



"""Terminal drawing module.

Draws data to terminal"""
from time import strftime
from typing import Any, Dict, List, Tuple

from rich.console import Console

from inky_pi.configs import STATION_FROM, STATION_TO
from inky_pi.display.display_base import DisplayBase, DisplayObject
from inky_pi.train.train_base import TrainBase
from inky_pi.weather.weather_base import IconType, ScaleType, WeatherBase


class TerminalDraw(DisplayBase):
    """Draw text and weather emoji in terminal window"""

    todo = "[bold red]TODO[/bold red]"

    def __init__(self, base_color: str = "") -> None:
        """Initialise terminal text library

        Args:
            base_color: base color
        """
        if base_color:
            self.todo = f"[bold {base_color}]{self.todo}[/bold {base_color}]"
        self._console = Console()
        self._output: List[str] = []

    def __enter__(self) -> "TerminalDraw":
        """Enter context manager

        Returns:
            self
        """
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit context manager"""
        self.render_text()

    def render_text(self) -> None:
        """Render collected text onto the terminal"""
        self._console.print("\n".join(self._output))

    def draw_date(self, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Append date to terminal text

        Args:
            x_y: (x, y) coordinates
        """
        date = strftime("%a %d %b %Y")
        self._output.append(date)

    def draw_time(self, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Append time to terminal text

        Args:
            x_y: (x, y) coordinates
        """
        time = strftime("%H:%M")
        self._output.append(time)

    def draw_train_times(
        self, data_t: TrainBase, num_trains: int = 3, x_y: Tuple[int, int] = (0, 0)
    ) -> None:
        """Append train data to terminal text

        Args:
            data_t: train data
            num_trains: number of trains to display
            x_y: (x, y) coordinates
        """
        self._output.append(f"train schedule from {STATION_FROM} to {STATION_TO}:")
        for i in range(0, num_trains):
            train = data_t.fetch_train(i)
            self._output.append(train)

    def draw_weather_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
        disp_tomorrow: bool = False,
    ) -> None:
        """Append weather forecast to terminal text

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
            disp_tomorrow: display tomorrow's forecast
        """
        current_temp = data_w.get_current_temperature(scale)
        current_condition = data_w.get_current_condition()
        temp_range = data_w.get_temp_range(0, scale)
        condition = data_w.get_condition(0)
        self._output.append(f"now: {current_temp}")
        self._output.append(f"now: {current_condition}")
        self._output.append(f"today: {temp_range}")
        self._output.append(f"today: {condition}")
        if disp_tomorrow:
            tomorrow_condition = data_w.get_condition(1)
            self._output.append(f"tomorrow: {tomorrow_condition}")

    def draw_mini_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
        day: int = 0,
    ) -> None:
        """Append mini weather forecast to terminal text

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
            day: day to display
        """
        self._output.append(f"{self.todo}: draw_mini_forecast")

    def draw_weather_icon(self, icon: IconType, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Append weather icon to terminal text

        Args:
            icon: icon type
            x_y: (x, y) coordinates
        """
        draw_icon_dispatcher: Dict[IconType, str] = {
            IconType.CLEAR_SKY: "\U00002600",
            IconType.FEW_CLOUDS: "\U000026C5",
            IconType.SCATTERED_CLOUDS: "\U000026C5",
            IconType.BROKEN_CLOUDS: "\U00002601",
            IconType.SHOWER_RAIN: "\U0001F327",
            IconType.RAIN: "\U0001F326",
            IconType.THUNDERSTORM: "\U000026C8",
            IconType.SNOW: "\U0001F328",
            IconType.MIST: "\U0001F32B",
        }
        emoji = draw_icon_dispatcher[icon]
        self._output.append(emoji)

    def draw_forecast_icons(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
    ) -> None:
        """Append extended weather forecast icons to terminal text

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
        """
        self._output.append(f"{self.todo}: draw_forecast_icons")

    def draw_goodnight(
        self, data_w: WeatherBase, scale: ScaleType = ScaleType.CELSIUS
    ) -> None:
        """Append goodnight message to terminal text

        Args:
            data_w: weather data
            scale: scale type
        """
        message = "Good Night ^^"
        temp = data_w.get_temp_range(1, scale)
        self._output.append(message)
        self._output.append(temp)


def instantiate_terminal_display(display_object: DisplayObject) -> TerminalDraw:
    """Terminal display object creator

    Args:
        display_object (DisplayObject): display object containing model

    Returns:
        TerminalDraw: TerminalDraw object
    """
    return TerminalDraw(display_object.base_color)



"""Tests for train module"""
import json
from pathlib import Path
from typing import Generator, Mapping, Union
from unittest.mock import Mock, patch

import pytest

from inky_pi.train.huxley2 import Huxley2
from inky_pi.train.open_live import OpenLive
from inky_pi.train.train_base import (
    TrainBase,
    TrainModel,
    TrainObject,
    abbreviate_stn_name,
)
from inky_pi.util import train_model_factory
from tests.unit.resources.fakes import FakeRequests

TEST_DIR = Path(__file__).parent
RESOURCES_DIR = TEST_DIR.joinpath("resources")
HUXLEY2_TRAIN_DATA = RESOURCES_DIR.joinpath("train_data.json")
OPEN_LIVE_TRAIN_DATA = RESOURCES_DIR.joinpath("train_data_zeep.pickle")
INVALID_OPEN_LIVE_DATA = RESOURCES_DIR.joinpath("trains_unavailable_zeep.pickle")


# pylint: disable=possibly-unused-variable
@pytest.fixture
def _setup_train_vars() -> Generator[Mapping[str, Union[int, float, str]], None, None]:
    station_from: str = "MZH"
    station_to: str = "LBG"
    number: int = 3
    url: str = "http://url.url"
    token: str = "key"
    yield locals()


@pytest.fixture
def _setup_train_object_open_live(
    _setup_train_vars: Mapping,
) -> Generator[TrainObject, None, None]:
    open_live_train_object = TrainObject(
        model=TrainModel.OPEN_LIVE,
        station_from=_setup_train_vars["station_from"],
        station_to=_setup_train_vars["station_to"],
        number=_setup_train_vars["number"],
        url=_setup_train_vars["url"],
        token=_setup_train_vars["token"],
    )
    yield open_live_train_object


@pytest.fixture
def _setup_train_object_huxley2(
    _setup_train_vars: Mapping,
) -> Generator[TrainObject, None, None]:
    huxley2_train_object = TrainObject(
        model=TrainModel.HUXLEY2,
        station_from=_setup_train_vars["station_from"],
        station_to=_setup_train_vars["station_to"],
        number=_setup_train_vars["number"],
    )
    yield huxley2_train_object


@pytest.fixture
def _setup_huxley2_fake_data(
    _setup_train_object_huxley2: TrainObject,
) -> Generator[Huxley2, None, None]:
    requests = FakeRequests()
    with open(HUXLEY2_TRAIN_DATA, "r", encoding="utf-8") as file:
        train_data = json.load(file)
        requests.add_response(train_data, 200)

    train_base = Huxley2()
    train_base.retrieve_data(requests, _setup_train_object_huxley2)
    yield train_base


@pytest.mark.parametrize(
    "name, expected_abbreviation",
    [
        ("London Cannon Street", "London Cannon St"),
        ("Clapham South", "Clapham S"),
        ("West Hampstead City Thameslink", "W Hampstead City TL"),
    ],
)
def test_abbreviate_station_name(name: str, expected_abbreviation: str) -> None:
    """Test for abbreviating station names

    Args:
        name (str): The name of the station
        expected_abbreviation (str): The expected abbreviation of the station
    """
    assert abbreviate_stn_name(name) == expected_abbreviation


def test_format_train_string() -> None:
    """Test for formatting train string"""
    arrival_t = "12:00"
    platform = "1"
    dest_stn = "London Cannon Street"
    status = "On time"
    assert TrainBase.format_train_string(arrival_t, platform, dest_stn, status) == (
        "12:00 | P1 to London Cannon St - On time"
    )


@pytest.mark.parametrize(
    "error_msg, num, expected",
    [
        ("Error message", 0, "Error message"),
        (
            "There are no train services at this station",
            0,
            "There are no train services at this st",
        ),
        ("Error message", 1, ""),
        (
            "There are no train services at station London Bridge",
            1,
            "London Bridge",
        ),
    ],
)
def test_format_train_error_string(error_msg: str, num: int, expected: str) -> None:
    """Test for formatting train error string

    Args:
        error_msg (str): The error message
        num (int): The number of trains
        expected (str): The expected string
    """
    assert TrainBase.format_error_msg(error_msg, num) == expected


@patch("inky_pi.train.open_live.zeep.plugins.HistoryPlugin")
@patch("inky_pi.train.open_live.zeep.xsd.Element")
@patch("inky_pi.train.open_live.zeep.Client")
def test_can_successfully_instantiate_train_open_live(
    zeep_client_mock: Mock,
    zeep_xsd_mock: Mock,
    zeep_history_mock: Mock,
    _setup_train_object_open_live: TrainObject,
) -> None:
    """Test for creating OpenLDBWS instanced object

    Args:
        zeep_client_mock (Mock): Mock for Client class
        zeep_xsd_mock (Mock): Mock for xsd.Element class
        zeep_history_mock (Mock): Mock for HistoryPlugin class
        _setup_train_object_open_live (TrainObject): Open Live TrainObject
    """
    ret: TrainBase = train_model_factory(_setup_train_object_open_live)
    zeep_history_mock.assert_called_once()
    zeep_client_mock.assert_called_once()
    assert zeep_xsd_mock.call_count == 2
    assert isinstance(ret, OpenLive)


@patch("inky_pi.train.huxley2.requests.get")
def test_can_successfully_instantiate_train_huxley2(
    requests_get_mock: Mock, _setup_train_object_huxley2: TrainObject
) -> None:
    """Test for creating Huxley2 OpenLDBWS instanced object

    Args:
        requests_get_mock (Mock): Mock for requests.get class
        _setup_train_object_huxley2 (TrainObject): Huxley2 TrainObject
    """
    ret: TrainBase = train_model_factory(_setup_train_object_huxley2)
    requests_get_mock.assert_called_once()
    assert isinstance(ret, Huxley2)


@patch("inky_pi.util.sys.exit")
def test_instantiate_open_live_without_url_and_token_raises_error(
    sys_exit_mock: Mock, _setup_train_vars: Mapping
) -> None:
    """Test for invalid OpenLDBWS object creation
    Due to the fact OpenLDBWS requires URL and token, test will trigger error

    Args:
        _setup_train_vars (Mapping): Mapping of variables to be used in the test
    """
    train_object = TrainObject(
        model=TrainModel.OPEN_LIVE,
        station_from=_setup_train_vars["station_from"],
        station_to=_setup_train_vars["station_to"],
        number=_setup_train_vars["number"],
    )
    with pytest.raises(ValueError):
        train_model_factory(train_object)
        sys_exit_mock.assert_called_once()


@pytest.mark.parametrize(
    "num, expected",
    [
        (0, "18:11 | P2 to Slade Green - On time"),
        (1, "18:14 | P1 to Luton - On time"),
        (2, "18:21 | P2 to Slade Green - On time"),
    ],
)
def test_can_successfully_fetch_train_string_huxley2(
    num: int, expected: str, _setup_huxley2_fake_data: Huxley2
) -> None:
    """Test for fetching train string from Huxley2 fake data

    Args:
        num (int): The number of trains to be fetched
        expected (str): The expected string
        _setup_huxley2_fake_data (Huxley2): Huxley2 fake data
    """
    assert _setup_huxley2_fake_data.fetch_train(num) == expected


@pytest.mark.parametrize(
    "num",
    [-1, 10],
)
def test_fetch_train_string_huxley2_invalid_num_raises_error(
    num: int,
    _setup_huxley2_fake_data: Huxley2,
) -> None:
    """Test invalid train number for Huxley2

    Args:
        num (int): The number of trains to be fetched
        _setup_huxley2_fake_data (Huxley2): Huxley2 fake data
    """
    with pytest.raises(ValueError):
        _setup_huxley2_fake_data.fetch_train(num)



"""Tests for utility methods and classes"""
from unittest.mock import Mock, patch

from inky_pi.util import LOG_FILE, LOG_ROTATION, LOG_SERIALIZE, configure_logging


@patch("inky_pi.util.logger.add")
def test_can_successfully_configure_logging(logger_add_mock: Mock) -> None:
    """Test for setting up logging"""
    configure_logging()
    logger_add_mock.assert_called_once_with(
        LOG_FILE, rotation=LOG_ROTATION, serialize=LOG_SERIALIZE
    )



"""Inky_Pi drawing module.

Draw strings and icons"""
import platform
from datetime import datetime, timedelta
from time import strftime
from typing import Any, Callable, Dict, Tuple

# pylint: disable=no-name-in-module
from font_fredoka_one import FredokaOne  # type: ignore
from font_hanken_grotesk import HankenGroteskBold  # type: ignore

# pylint: enable=no-name-in-module
from PIL import Image, ImageDraw, ImageFont  # type: ignore

from inky_pi.display.display_base import DisplayBase, DisplayModel, DisplayObject
from inky_pi.display.util.desktop_driver import DesktopDisplayDriver
from inky_pi.display.util.drawing import (
    draw_cloud_icon,
    draw_cloud_lightning_icon,
    draw_cloud_rain_icon,
    draw_cloud_snow_icon,
    draw_mist_icon,
    draw_sun_cloud_icon,
    draw_sun_cloud_rain_icon,
    draw_sun_icon,
    draw_two_clouds_icon,
)
from inky_pi.display.util.shapes import gen_closed_eye_icon
from inky_pi.train.train_base import TrainBase
from inky_pi.weather.weather_base import IconType, ScaleType, WeatherBase

# Font constants
FONT_XS = ImageFont.truetype(HankenGroteskBold, 16)
FONT_S = ImageFont.truetype(HankenGroteskBold, 20)
FONT_M = ImageFont.truetype(HankenGroteskBold, 25)
FONT_L = ImageFont.truetype(HankenGroteskBold, 35)
FONT_XL = ImageFont.truetype(HankenGroteskBold, 40)
FONT_GS = ImageFont.truetype(FredokaOne, 25)
FONT_GM = ImageFont.truetype(FredokaOne, 30)
FONT_GL = ImageFont.truetype(FredokaOne, 40)


class InkyDraw(DisplayBase):
    """Draw text and shapes onto Inky e-ink display"""

    def __init__(self, display_driver: Any) -> None:
        """Create display and image drawing objects
        inky_model can be an InkyWHAT model or a DesktopDisplayDriver object

        Args:
            display_driver (Any): Display driver (InkyWHAT or DesktopDisplayDriver)
        """
        self._display: Any = display_driver
        self._img: Image = Image.new(
            "P", (self._display.WIDTH, self._display.HEIGHT), color="white"
        )
        self._img_draw: ImageDraw = ImageDraw.Draw(self._img)
        self._black: Any = self._display.BLACK
        self._white: Any = self._display.WHITE
        self._color: Any = self._display.YELLOW

    def __enter__(self) -> "InkyDraw":
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit context manager; render the drawn screen"""
        self.render_screen()

    def render_screen(self) -> None:
        """Render border, images (w/text) on inky screen and show on display"""
        self._display.set_image(self._img)
        self._display.set_border(self._black)
        self._display.show()

    def draw_goodnight(
        self, data_w: WeatherBase, scale: ScaleType = ScaleType.CELSIUS
    ) -> None:
        """Render goodnight screen

        Args:
            data_w (WeatherBase): Weather data object
            scale (ScaleType): Scale type
        """
        x_mid, y_mid = self._display.WIDTH / 2, self._display.HEIGHT / 2
        gen_closed_eye_icon(self._img_draw, self._color, (x_mid, y_mid))
        # Message text
        message_str = "Good Night ^^"
        width, height = FONT_GL.getsize(message_str)
        message_x, message_y = x_mid - (width / 2), y_mid - (height / 2)
        self._img_draw.text((message_x, message_y), message_str, self._black, FONT_GL)
        # Weather text
        x_weather, y_weather = 20, 210
        self._img_draw.text(
            (x_weather, y_weather),
            data_w.get_temp_range(1, scale),
            self._color,
            FONT_GM,
        )
        self._img_draw.text(
            (x_weather, y_weather + 40), data_w.get_condition(1), self._color, FONT_GS
        )

    def draw_date(self, x_y: Tuple[int, int] = (10, 5)) -> None:
        """Draw date text

        Args:
            x_y: (x, y) coordinates
        """
        self._img_draw.text(x_y, strftime("%a %d %b %Y"), self._black, FONT_S)

    def draw_time(self, x_y: Tuple[int, int] = (257, 5)) -> None:
        """Draw time text

        Args:
            x_y: (x, y) coordinates
        """
        self._img_draw.text(x_y, f"Updated {strftime('%H:%M')}", self._black, FONT_S)

    def draw_train_times(
        self, data_t: TrainBase, num_trains: int = 3, x_y: Tuple[int, int] = (10, 205)
    ) -> None:
        """Draw all train times text

        Each line: Train time, platform, destination station, ETA

        Args:
            data_t (TrainBase): TrainBase object
            num_trains (int): Number of train info to draw
            x_y: (x, y) coordinates
        """
        for i in range(0, num_trains):
            self._img_draw.text(
                (x_y[0], x_y[1] + i * 30),
                data_t.fetch_train(i),
                self._black,
                FONT_S,
            )

    def draw_weather_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (135, 50),
        disp_tomorrow: bool = False,
    ) -> None:
        """Draw all weather forecast text

        First line: current temperature and weather condition
        Second line: today's temperature range
        Third line: today's weather condition description
        Fourth line: tomorrow's weather condition description

        Args:
            data_w (WeatherBase): WeatherBase object
            scale (ScaleType): Celsius or Fahrenheit for formatting
            x_y: (x, y) coordinates
            disp_tomorrow (bool): Display tomorrow's weather forecast
        """
        self._img_draw.text(
            x_y,
            data_w.get_current_temperature(scale),
            self._black,
            FONT_XL,
        )
        self._img_draw.text(
            (x_y[0] + 140, x_y[1] + 13),
            data_w.get_current_condition(),
            self._black,
            FONT_M,
        )
        self._img_draw.text(
            (x_y[0], x_y[1] + 50),
            data_w.get_temp_range(0, scale),
            self._black,
            FONT_M,
        )
        self._img_draw.text(
            (x_y[0], x_y[1] + 80), data_w.get_condition(0), self._black, FONT_M
        )
        if disp_tomorrow:
            self._img_draw.text(
                (x_y[0], x_y[1] + 110),
                "tomorrow: " + data_w.get_condition(1),
                self._black,
                FONT_S,
            )

    def draw_weather_icon(
        self,
        icon: IconType,
        x_y: Tuple[int, int] = (30, 90),
    ) -> None:
        """Draws specified icon

        Args:
            icon (IconType): Weather IconType to draw
            x_y: (x, y) coordinates
        """
        draw_icon_dispatcher: Dict[IconType, Callable] = {
            IconType.CLEAR_SKY: draw_sun_icon,
            IconType.FEW_CLOUDS: draw_sun_cloud_icon,
            IconType.SCATTERED_CLOUDS: draw_cloud_icon,
            IconType.BROKEN_CLOUDS: draw_two_clouds_icon,
            IconType.SHOWER_RAIN: draw_cloud_rain_icon,
            IconType.RAIN: draw_sun_cloud_rain_icon,
            IconType.THUNDERSTORM: draw_cloud_lightning_icon,
            IconType.SNOW: draw_cloud_snow_icon,
            IconType.MIST: draw_mist_icon,
        }
        draw_icon_dispatcher[icon](self._img_draw, self._black, self._white, x_y)

    def draw_mini_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (30, 40),
        day: int = 0,
    ) -> None:
        """Draws weather forecast icons and text future day

        Args:
            data_w (WeatherBase): WeatherBase object
            scale (ScaleType): Celsius or Fahrenheit for formatting
            x_y: (x, y) coordinates
            day (int): Day to display
        """
        new_date = datetime.now() + timedelta(days=day)
        if day > 0:
            self._img_draw.text(
                (x_y[0] + 10, x_y[1] + 5),
                new_date.strftime("%a %d"),
                self._black,
                FONT_XS,
            )
        self.draw_weather_icon(data_w.get_icon(day), (x_y[0], x_y[1] + 27))
        self._img_draw.text(
            ((x_y[0] + 10 if day > 0 else x_y[0]), x_y[1] + 90),
            data_w.get_future_weather(day, scale),
            self._black,
            FONT_XS if day > 0 else FONT_S,
        )

    def draw_forecast_icons(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (10, 180),
    ) -> None:
        """Draws weather forecast icons and text for next 5 days

        Args:
            data_w (WeatherBase): WeatherBase object
            scale (ScaleType): Celsius or Fahrenheit for formatting
            x_y: (x, y) coordinates
        """
        spacing: int = 78
        for i in range(0, 5):
            self.draw_mini_forecast(
                data_w, scale, (x_y[0] + (i * spacing), x_y[1]), i + 1
            )


def _import_inky_what() -> Any:
    """This is a wrapper that imports InkyWHAT to allow mocking it in tests.

    Returns:
        InkyWHAT: Raspberry pi InkyWHAT library
    """
    if platform.machine() == "armv7l":
        # pylint: disable=import-outside-toplevel
        from inky import InkyWHAT  # type: ignore

        return InkyWHAT
    raise ImportError("InkyWHAT library unavailable (are you on a Raspberry Pi?)")


def instantiate_inky_display(display_object: DisplayObject) -> InkyDraw:
    """Inky display object creator

    Args:
        display_object (DisplayObject): display object containing model

    Returns:
        InkyDraw: InkyDraw object
    """
    display_driver = (
        _import_inky_what()
        if display_object.model == DisplayModel.INKY_WHAT
        else DesktopDisplayDriver
    )
    return InkyDraw(display_driver(f"{display_object.base_color}"))



"""Utility functions for inky_pi."""
import sys
from pathlib import Path
from typing import Callable, Dict

from loguru import logger

from inky_pi.display.display_base import DisplayBase, DisplayModel, DisplayObject
from inky_pi.display.inky_draw import instantiate_inky_display
from inky_pi.display.terminal_draw import instantiate_terminal_display
from inky_pi.train.huxley2 import instantiate_huxley2
from inky_pi.train.open_live import instantiate_open_live
from inky_pi.train.train_base import TrainBase, TrainModel, TrainObject
from inky_pi.weather.open_weather_map import instantiate_open_weather_map
from inky_pi.weather.weather_base import WeatherBase, WeatherModel, WeatherObject

LOG_ROOT_DIR = Path(__file__).parent.parent
LOG_FILE = LOG_ROOT_DIR.joinpath("inky.log")
LOG_ROTATION = "5 MB"
LOG_SERIALIZE = True


def configure_logging() -> None:
    """Configure logging options

    See: https://loguru.readthedocs.io/en/stable/api.html
    """
    logger.add(LOG_FILE, rotation=LOG_ROTATION, serialize=LOG_SERIALIZE)


def import_display(display_object: DisplayObject) -> DisplayBase:
    """
    Imports the display model from the given display object and returns it.
    On error, logs and exits program (typically if not running on RPi).

    Args:
        display_object (DisplayObject): The display object to be imported.

    Returns:
        DisplayBase: The display model.
    """
    try:
        return display_model_factory(display_object)
    except ImportError as exc:
        logger.error(exc)
        sys.exit(1)


def display_model_factory(display_object: DisplayObject) -> DisplayBase:
    """Selects and instantiates the defined display model to use

    Args:
        display_object (DisplayObject): display object containing model

    Returns:
        DisplayBase: DisplayBase object
    """
    display_handler: Dict[DisplayModel, Callable[[DisplayObject], DisplayBase]] = {
        DisplayModel.INKY_WHAT: instantiate_inky_display,
        DisplayModel.TERMINAL: instantiate_terminal_display,
        DisplayModel.DESKTOP: instantiate_inky_display,
    }
    return display_handler[display_object.model](display_object)


def _check_open_live_params(train_object: TrainObject) -> None:
    """Checks if the Open Live URL and API token have been provided

    Args:
        train_object (TrainObject): train object containing model

    Raises:
        ValueError: If the Open Live API token is invalid
    """
    if train_object.model != TrainModel.OPEN_LIVE:
        return
    if train_object.url == "" or train_object.token == "":
        raise ValueError("Open Live requires URL and API token.")


def train_model_factory(train_object: TrainObject) -> TrainBase:
    """Selects and instantiates the defined train model to use

    Args:
        train_object (TrainObject): train object containing model

    Returns:
        TrainBase: TrainBase object
    """
    _check_open_live_params(train_object)
    train_handler: Dict[TrainModel, Callable[[TrainObject], TrainBase]] = {
        TrainModel.OPEN_LIVE: instantiate_open_live,
        TrainModel.HUXLEY2: instantiate_huxley2,
    }
    try:
        return train_handler[train_object.model](train_object)
    except ValueError as exc:
        logger.error(exc)
        sys.exit(1)


def weather_model_factory(weather_object: WeatherObject) -> WeatherBase:
    """Selects and instantiates the defined weather model to use

    Args:
        weather_object (WeatherObject): weather object containing model

    Returns:
        WeatherBase: WeatherBase object
    """
    weather_handler: Dict[WeatherModel, Callable[[WeatherObject], WeatherBase]] = {
        WeatherModel.OPEN_WEATHER_MAP: instantiate_open_weather_map,
    }
    try:
        return weather_handler[weather_object.model](weather_object)
    except ValueError as exc:
        logger.error(exc)
        sys.exit(1)



"""Open Live Departure Boards Web Service (OpenLDBWS) API
"""
from typing import Any

import zeep
from loguru import logger

from inky_pi.train.train_base import TrainBase, TrainObject, abbreviate_stn_name


class OpenLive(TrainBase):
    """Fetch and manage train data"""

    def retrieve_data(self, protocol: Any, train_object: TrainObject) -> None:
        """Requests train data from OpenLDBWS train arrivals API endpoint

        API description: http://lite.realtime.nationalrail.co.uk/openldbws/

        Args:
            protocol (Any): Zeep object for SOAP requests
            train_object (TrainObject): Train object
        """
        history: Any = protocol.plugins.HistoryPlugin()
        client: Any = protocol.Client(wsdl=train_object.url, plugins=[history])
        header: Any = protocol.xsd.Element(
            "{http://thalesgroup.com/RTTI/2013-11-28/Token/types}AccessToken",
            protocol.xsd.ComplexType(
                [
                    protocol.xsd.Element(
                        "{http://thalesgroup.com/RTTI/2013-11-28/Token/types}"
                        + "TokenValue",
                        protocol.xsd.String(),
                    ),
                ]
            ),
        )
        header_value = header(TokenValue=train_object.token)
        self._num = train_object.number
        try:
            self._data = client.service.GetDepartureBoard(
                numRows=train_object.number,
                crs=train_object.station_from,
                filterCrs=train_object.station_to,
                filterType="to",
                _soapheaders=[header_value],
            )
            self.origin = abbreviate_stn_name(self._data.locationName)
            self.destination = abbreviate_stn_name(self._data.filterLocationName)
        except protocol.exceptions.Fault as exc:
            logger.error("Error retrieving train data (check stations?).")
            raise ValueError(f"Invalid train data request: {train_object}") from exc

    def fetch_train(self, num: int) -> str:
        """Generate next train string

        String is returned in format:
            [hh:mm] | [Platform #] to [Final Destination Station] - [Status]

        Args:
            num (int): Next train departing number starting from 0

        Returns:
            str: Formatted string or error message
        """
        self._validate_number(num)
        if not self._data:
            raise ValueError("No train data available.")

        try:
            service: Any = self._data.trainServices.service[num]
            platform: str = service.platform[0:2]
            arrival_t: str = service.std
            dest_stn: str = service.destination.location[0].locationName
            status: str = service.etd
            return TrainBase.format_train_string(arrival_t, platform, dest_stn, status)
        except (AttributeError, TypeError, KeyError, IndexError):
            return self._handle_error(num)

    def _handle_error(self, num: int) -> str:
        """Log error and raise exception

        Args:
            num (int): Train number

        Raises:
            Exception: Exception to raise
        """
        if not self._data:
            raise ValueError("No train data available.")

        try:
            # pylint: disable=protected-access
            error_msg = str(self._data.nrccMessages.message[0]._value_1)[1:]
            return TrainBase.format_error_msg(error_msg, num)
        except (AttributeError, TypeError, KeyError, IndexError):
            error_msg = f"No trains to {self.destination} from {self.origin}."
            return TrainBase.format_error_msg(error_msg, num)


def instantiate_open_live(train_object: TrainObject) -> OpenLive:
    """Open Live object creator

    Args:
        train_object (TrainObject): train object containing model

    Returns:
        OpenLive: OpenLive object
    """
    train_base = OpenLive()
    train_base.retrieve_data(zeep, train_object)
    return train_base



"""Inky Pi base display module.

Abstract class for all display models"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Tuple

from inky_pi.train.train_base import TrainBase
from inky_pi.weather.weather_base import IconType, ScaleType, WeatherBase


class DisplayModel(Enum):
    """Enum of display models"""

    INKY_WHAT = auto()
    TERMINAL = auto()
    DESKTOP = auto()


@dataclass
class DisplayObject:
    """Display object"""

    model: DisplayModel
    base_color: str = ""


class DisplayBase(ABC):
    """Abstract base class for all display models"""

    @abstractmethod
    def __enter__(self) -> "DisplayBase":
        """Enter context manager

        Returns:
            self
        """

    @abstractmethod
    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit context manager"""

    @abstractmethod
    def draw_date(self, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Display date

        Args:
            x_y: (x, y) coordinates
        """

    @abstractmethod
    def draw_time(self, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Display time

        Args:
            x_y: (x, y) coordinates
        """

    @abstractmethod
    def draw_train_times(
        self, data_t: TrainBase, num_trains: int = 0, x_y: Tuple[int, int] = (0, 0)
    ) -> None:
        """Display train data

        Args:
            data_t: train data
            num_trains: number of trains to display
            x_y: (x, y) coordinates
        """

    @abstractmethod
    def draw_weather_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
        disp_tomorrow: bool = False,
    ) -> None:
        """Display weather forecast

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
            disp_tomorrow: display tomorrow's forecast
        """

    @abstractmethod
    def draw_mini_forecast(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
        day: int = 0,
    ) -> None:
        """Display mini weather forecast

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
            day: day to display
        """

    @abstractmethod
    def draw_weather_icon(self, icon: IconType, x_y: Tuple[int, int] = (0, 0)) -> None:
        """Display weather icon

        Args:
            icon: icon type
            x_y: (x, y) coordinates
        """

    @abstractmethod
    def draw_forecast_icons(
        self,
        data_w: WeatherBase,
        scale: ScaleType = ScaleType.CELSIUS,
        x_y: Tuple[int, int] = (0, 0),
    ) -> None:
        """Display extended forecast icons

        Args:
            data_w: weather data
            scale: scale type
            x_y: (x, y) coordinates
        """

    @abstractmethod
    def draw_goodnight(
        self, data_w: WeatherBase, scale: ScaleType = ScaleType.CELSIUS
    ) -> None:
        """Display goodnight message

        Args:
            data_w: weather data
            scale: scale type
        """



"""Inky_Pi train model module.

Fetches train data from Huxley2 (OpenLDBWS) and generates formatted data"""
from typing import Any, Dict

import requests
from loguru import logger

from inky_pi.train.train_base import TrainBase, TrainObject, abbreviate_stn_name


class Huxley2(TrainBase):
    """Fetch and manage train data"""

    def retrieve_data(self, protocol: Any, train_object: TrainObject) -> None:
        """Requests train data from OpenLDBWS train arrivals API endpoint

        More info here: https://huxley2.azurewebsites.net/

        Args:
            protocol (Any): Requests object for HTTP requests
            train_object (TrainObject): Train object
        """
        response: Any = protocol.get(
            "https://huxley2.azurewebsites.net/departures/"
            f"{train_object.station_from}/to/"
            f"{train_object.station_to}/{train_object.number}"
        )

        self._num = train_object.number
        try:
            self._data = response.json()
            self.origin = abbreviate_stn_name(self._data["locationName"])
            self.destination = abbreviate_stn_name(self._data["filterLocationName"])
        except protocol.exceptions.JSONDecodeError as exc:
            logger.error("Error retrieving train data (check stations?).")
            raise ValueError(f"Invalid train data request: {train_object}") from exc

    def fetch_train(self, num: int) -> str:
        """Generate next train string

        String is returned in format:
            [hh:mm] | [Platform #] to [Final Destination Station] - [Status]

        Args:
            num (int): Next train departing number starting from 0

        Returns:
            str: Formatted string or error message
        """
        self._validate_number(num)
        if not self._data:
            raise ValueError("No train data available.")

        try:
            service: Dict = self._data["trainServices"][num]
            platform: str = service["platform"][0:2]
            arrival_t: str = service["std"]
            dest_stn: str = service["destination"][0]["locationName"]
            status: str = service["etd"]
            return TrainBase.format_train_string(arrival_t, platform, dest_stn, status)
        except (KeyError, TypeError, IndexError):
            return self._handle_error(num)

    def _handle_error(self, num: int) -> str:
        """Log error and raise exception

        Args:
            num (int): Train number

        Raises:
            Exception: Exception to raise
        """
        if not self._data:
            raise ValueError("No train data available.")

        try:
            error_msg = str(self._data["nrccMessages"][0]["value"])
            return TrainBase.format_error_msg(error_msg, num)
        except (AttributeError, TypeError, KeyError, IndexError):
            error_msg = f"No trains to {self.destination} from {self.origin}."
            return TrainBase.format_error_msg(error_msg, num)


def instantiate_huxley2(train_object: TrainObject) -> Huxley2:
    """Huxley2 object creator

    Args:
        train_object (TrainObject): train object containing model

    Returns:
        Huxley2: Huxley2 object
    """
    train_base = Huxley2()
    train_base.retrieve_data(requests, train_object)
    return train_base



"""Inky_Pi configuration settings."""
from environs import Env

_env: Env = Env()
_env.read_env()

TRAIN_MODEL: str = _env.str("TRAIN_MODEL", default="OPEN_LIVE")
"""Train model to use (OPEN_LIVE or HUXLEY2)"""

STATION_FROM: str = _env.str("STATION_FROM", default="BHO")
"""From (departing) station"""

STATION_TO: str = _env.str("STATION_TO", default="WMW")
"""To (arrival) station"""

TRAIN_NUMBER: int = _env.int("TRAIN_NUMBER", default=3)
"""Number of departing trains to fetch"""

TRAIN_API_TOKEN: str = _env.str("TRAIN_API_TOKEN", default="keep-in-.env-file")
"""OpenLDBWS API Token
Register: https://realtime.nationalrail.co.uk/OpenLDBWSRegistration/Registration
Keep this in a .env file in project root dir i.e.: TRAIN_API_TOKEN=keyvalue
"""

TRAIN_MODEL_URL: str = _env.str(
    "TRAIN_MODEL_URL",
    default="http://lite.realtime.nationalrail.co.uk/"
    + "OpenLDBWS/wsdl.aspx?ver=2017-10-01",
)
"""OpenLDBWS API Endpoint"""

WEATHER_MODEL: str = _env.str("WEATHER_MODEL", default="OPEN_WEATHER_MAP")
"""Weather model to use (OPEN_WEATHER_MAP only for now)"""

LATITUDE: float = _env.float("LATITUDE", default=51.5085)
"""Target Latitude (default in London)"""

LONGITUDE: float = _env.float("LONGITUDE", default=-0.1257)
"""Target Longitude (default in London)"""

EXCLUDE_FLAGS: str = _env.str("EXCLUDE_FLAGS", default="minutely,hourly")
"""Weather data exclude flags"""

WEATHER_API_TOKEN: str = _env.str("WEATHER_API_TOKEN", default="keep-in-.env-file")
"""OpenWeatherMap API Key
Keys: https://home.openweathermap.org/api_keys
Keep this in a .env file in project root dir i.e.: WEATHER_API_TOKEN=keyvalue
"""



"""Inky_Pi main module.

Fetches Train and Weather data and displays on a Raspberry Pi w/InkyWHAT."""
from argparse import ArgumentParser, Namespace
from enum import Enum, auto
from typing import Dict

from loguru import logger

from inky_pi.__init__ import __version__  # type: ignore
from inky_pi.configs import (
    EXCLUDE_FLAGS,
    LATITUDE,
    LONGITUDE,
    STATION_FROM,
    STATION_TO,
    TRAIN_API_TOKEN,
    TRAIN_MODEL,
    TRAIN_MODEL_URL,
    TRAIN_NUMBER,
    WEATHER_API_TOKEN,
    WEATHER_MODEL,
)
from inky_pi.display.display_base import DisplayModel, DisplayObject
from inky_pi.train.train_base import TrainBase, TrainModel, TrainObject
from inky_pi.util import (
    configure_logging,
    import_display,
    train_model_factory,
    weather_model_factory,
)
from inky_pi.weather.weather_base import (
    ScaleType,
    WeatherBase,
    WeatherModel,
    WeatherObject,
)

# Define objects to be used in fetching data
TRAIN_OBJECT = TrainObject(
    model=TrainModel[TRAIN_MODEL],
    station_from=STATION_FROM,
    station_to=STATION_TO,
    number=TRAIN_NUMBER,
    url=TRAIN_MODEL_URL,
    token=TRAIN_API_TOKEN,
)

WEATHER_OBJECT = WeatherObject(
    model=WeatherModel[WEATHER_MODEL],
    latitude=LATITUDE,
    longitude=LONGITUDE,
    exclude_flags=EXCLUDE_FLAGS,
    weather_api_token=WEATHER_API_TOKEN,
)


OUTPUT_HANDLER: Dict[str, DisplayObject] = {
    "inky": DisplayObject(model=DisplayModel.INKY_WHAT, base_color="black"),
    "terminal": DisplayObject(model=DisplayModel.TERMINAL),
    "desktop": DisplayObject(model=DisplayModel.DESKTOP),
}


class DisplayOption(Enum):
    """Enum of display options"""

    TRAIN = auto()
    WEATHER = auto()
    NIGHT = auto()


def main() -> None:
    """The entry point for the program. It parses the command line arguments and calls
    the appropriate display function and output option based on the arguments.
    """
    configure_logging()
    logger.debug("InkyPi main initialized")
    args: Namespace = _parse_args()
    _create_display(args)


def _parse_args() -> Namespace:
    """Parses the command line arguments and returns the parsed arguments.

    Returns:
        Namespace: Parsed command line arguments.
    """
    parser = ArgumentParser(
        description="Displays train and weather data to various outputs",
    )
    parser.add_argument(
        "-o",
        "--option",
        help="Display option (train, weather, night)",
        type=str,
        default="train",
    )
    parser.add_argument(
        "-m",
        "--output",
        help="Output source (inky, terminal, desktop)",
        type=str,
        default="inky",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Inky_Pi version",
        action="version",
        version="%(prog)s " + __version__,
    )
    return parser.parse_args()


def _create_display(args: Namespace) -> None:
    """Takes a Namespace object and uses the display and output attributes

    If the display or output attributes are not found in the dispatchers, the
    program will log the error and exit.

    Args:
        args (Namespace): Parsed command line arguments
    """
    try:
        display_data(
            DisplayOption[args.option.upper()], OUTPUT_HANDLER[args.output.lower()]
        )
    except KeyError:
        logger.error(
            f'Invalid display/output specified: "{args.display}"/"{args.output}"'
        )
    except Exception as exc:  # pylint: disable=broad-except
        logger.exception(exc)


def display_data(option: DisplayOption, output: DisplayObject) -> None:
    """inky_pi weather with train function

    Retrieves train and weather data from API endpoints, generates text and
    weather icon, and draws to inkyWHAT screen.

    Args:
        option (DisplayOption): Display option (train, weather, night)
        output (DisplayObject): Display object (inkyWHAT, terminal, desktop)
    """
    output.base_color = "yellow" if option == DisplayOption.NIGHT else "black"

    weather_data: WeatherBase = weather_model_factory(WEATHER_OBJECT)
    with import_display(output) as display:
        if option == DisplayOption.NIGHT:
            display.draw_goodnight(weather_data)
            return

        display.draw_date()
        display.draw_time()
        display.draw_weather_icon(weather_data.get_icon())
        display.draw_weather_forecast(
            weather_data,
            ScaleType.CELSIUS,
            disp_tomorrow=bool(option == DisplayOption.TRAIN),
        )
        if option == DisplayOption.TRAIN:
            train_data: TrainBase = train_model_factory(TRAIN_OBJECT)
            display.draw_train_times(train_data, TRAIN_NUMBER)
        elif option == DisplayOption.WEATHER:
            display.draw_forecast_icons(weather_data)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inky_pi',
 'inky_pi.display',
 'inky_pi.display.util',
 'inky_pi.train',
 'inky_pi.weather',
 'tests',
 'tests.e2e',
 'tests.unit',
 'tests.unit.resources']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.0,<10.0.0',
 'click>=8.1.2,<9.0.0',
 'environs>=9.3.2,<10.0.0',
 'font-fredoka-one>=0.0.4,<0.0.5',
 'font-hanken-grotesk>=0.0.2,<0.0.3',
 'loguru>=0.5.3,<0.6.0',
 'numpy==1.21.1',
 'rich>=12.4.1,<13.0.0',
 'urllib3>=1.26.8,<2.0.0',
 'zeep>=4.0.0,<5.0.0']

extras_require = \
{':platform_machine == "armv7l"': ['RPi.GPIO>=0.7.0,<0.8.0',
                                   'inky>=1.2.0,<2.0.0']}

entry_points = \
{'console_scripts': ['inky_pi = inky_pi.cli:main']}

setup_kwargs = {
    'name': 'inky-pi',
    'version': '0.1.0',
    'description': 'Top-level package for inky-pi.',
    'long_description': '=======\ninky-pi\n=======\n\n\n.. image:: https://img.shields.io/pypi/v/inky_pi.svg\n        :target: https://pypi.python.org/pypi/inky_pi\n\n.. image:: https://github.com/mickeykkim/inky_pi/actions/workflows/main.yml/badge.svg\n        :target: https://github.com/mickeykkim/inky_pi/actions/workflows/main.yml\n\n.. image:: https://readthedocs.org/projects/inky-pi/badge/?version=latest\n        :target: https://inky-pi.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\nThis program generates weather and UK train data for output to various displays (inkyWHAT e-ink display, terminal, desktop png).\n\n\n* Free software: MIT\n* Documentation: https://inky-pi.readthedocs.io.\n\n\nFeatures\n--------\n\n* Displays weather data using Weather Underground API\n* Displays UK train data using National Rail API\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Mickey Kim',
    'author_email': 'mickeykkim@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mickeykkim/inky_pi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



"""Tests for weather module"""
import json
from math import isclose
from pathlib import Path
from typing import Generator
from unittest.mock import Mock, patch

import pytest

from inky_pi.util import weather_model_factory
from inky_pi.weather.open_weather_map import DEG_C, DEG_F, OpenWeatherMap
from inky_pi.weather.weather_base import (
    IconType,
    ScaleType,
    WeatherBase,
    WeatherModel,
    WeatherObject,
    celsius_to_fahrenheit,
    kelvin_to_celsius,
)
from tests.unit.resources.fakes import FakeRequests

TEST_DIR = Path(__file__).parent
RESOURCES_DIR = TEST_DIR.joinpath("resources")
WEATHER_DATA = RESOURCES_DIR.joinpath("weather_data.json")
INVALID_WEATHER_DATA = RESOURCES_DIR.joinpath("weather_cod401.json")


# pylint: disable=possibly-unused-variable
@pytest.fixture
def _setup_weather_object() -> Generator[WeatherObject, None, None]:
    weather_object = WeatherObject(
        model=WeatherModel.OPEN_WEATHER_MAP,
        latitude=51.5085,
        longitude=-0.1257,
        exclude_flags="minutely,hourly",
        weather_api_token="1234567890",
    )
    yield weather_object


@pytest.fixture
def _setup_weather_fake_data(
    _setup_weather_object: WeatherObject,
) -> Generator[OpenWeatherMap, None, None]:
    requests = FakeRequests()
    with open(WEATHER_DATA, "r", encoding="utf-8") as file:
        weather_data = json.load(file)
        requests.add_response(weather_data, 200)

    weather_base = OpenWeatherMap()
    weather_base.retrieve_data(requests, _setup_weather_object)
    yield weather_base


@pytest.mark.parametrize(
    "celsius, fahrenheit",
    [
        (-51, -59.8),
        (0, 32.0),
        (100, 212.0),
    ],
)
def test_celsius_to_fahrenheit(celsius: float, fahrenheit: float) -> None:
    """Test for converting celsius to fahrenheit
    Conversion significant to tenths place

    Args:
        celsius (float): Celsius value
        fahrenheit (float): Expected fahrenheit value
    """
    assert isclose(celsius_to_fahrenheit(celsius), fahrenheit)


def test_celsius_below_abs_zero_raises_value_error() -> None:
    """Rests that the function `celsius_to_fahrenheit` raises a `ValueError`
    when given a temperature below absolute zero
    """
    with pytest.raises(ValueError):
        celsius_to_fahrenheit(-273.16)


@pytest.mark.parametrize(
    "kelvin, celsius",
    [
        (0, -273.1),
        (100, -173.1),
        (373.15, 100.0),
    ],
)
def test_kelvin_to_celsius(kelvin: float, celsius: float) -> None:
    """Test for converting kelvin to celsius
    Conversion significant to tenths place

    Args:
        kelvin (float): Kelvin value
        celsius (float): Expected celsius value
    """
    assert isclose(kelvin_to_celsius(kelvin), celsius)


def test_kelvin_below_abs_zero_raises_value_error() -> None:
    """Rests that the `kelvin_to_celsius` function raises a `ValueError`
    when given a temperature below absolute zero
    """
    with pytest.raises(ValueError):
        kelvin_to_celsius(-1)


@patch("inky_pi.weather.open_weather_map.requests.get")
def test_can_successfully_instantiate_weather_open_weather_map(
    requests_get_mock: Mock, _setup_weather_object: WeatherObject
) -> None:
    """Test for creating OpenWeatherMap instanced object

    Args:
        requests_get_mock (Mock): Mock for requests.get
        _setup_weather_object (WeatherObject): Setup weather data
    """
    weather_object = _setup_weather_object
    ret: WeatherBase = weather_model_factory(weather_object)
    requests_get_mock.assert_called_once()
    assert isinstance(ret, OpenWeatherMap)


def test_retrieving_invalid_weather_data_raises_value_error(
    _setup_weather_object: WeatherObject,
) -> None:
    """Test to detect sending an invalid API key response

    Args:
        _setup_weather_object (WeatherObject): Setup weather data
    """
    _invalid_requests = FakeRequests()
    with open(INVALID_WEATHER_DATA, "r", encoding="utf-8") as file:
        weather_data = json.load(file)
        _invalid_requests.add_response(weather_data, 200)

    with pytest.raises(ValueError) as exc_info:
        weather_base = OpenWeatherMap()
        weather_base.retrieve_data(_invalid_requests, _setup_weather_object)
        assert "Invalid API Key" in str(exc_info.value)


@pytest.mark.parametrize(
    "day, expected_icon_type",
    [
        (0, IconType.SCATTERED_CLOUDS),
        (1, IconType.CLEAR_SKY),
        (2, IconType.RAIN),
        (3, IconType.BROKEN_CLOUDS),
        (4, IconType.SHOWER_RAIN),
        (5, IconType.THUNDERSTORM),
        (6, IconType.SNOW),
        (7, IconType.MIST),
    ],
)
def test_can_successfully_retrieve_weather_icon(
    day: int, expected_icon_type: IconType, _setup_weather_fake_data: OpenWeatherMap
) -> None:
    """Test for retrieving weather icons

    Args:
        day (int): Day of the week
        expected_icon_type (IconType): Expected icon type
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    assert weather_obj.get_icon(day) == expected_icon_type


@pytest.mark.parametrize(
    "day, condition",
    [
        (0, "light rain"),
        (1, "clear sky"),
        (2, "heavy intensity rain"),
    ],
)
def test_can_successfully_retrieve_condition(
    day: int, condition: str, _setup_weather_fake_data: OpenWeatherMap
) -> None:
    """Test for retrieving weather conditions

    Args:
        day (int): Day of the week
        condition (str): Expected condition
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    assert weather_obj.get_condition(day) == condition


@pytest.mark.parametrize(
    "day",
    [-1, 8],
)
def test_retrieving_invalid_day_condition_raises_value_error(
    day: int, _setup_weather_fake_data: OpenWeatherMap
) -> None:
    """Test for that retrieving invalid day condition raises ValueError

    Args:
        day (int): Day of the week
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    with pytest.raises(ValueError):
        weather_obj.get_condition(day)


def test_can_successfully_retrieve_current_condition(
    _setup_weather_fake_data: OpenWeatherMap,
) -> None:
    """Test for retrieving current condition

    Args:
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    current_condition = "Clouds"
    assert weather_obj.get_current_condition() == current_condition


def test_can_successfully_retrieve_current_weather(
    _setup_weather_fake_data: OpenWeatherMap,
) -> None:
    """Test for retrieving current weather

    Args:
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    current_temp_k = 289.46
    current_condition = "Clouds"
    expected_string_c = (
        f"{kelvin_to_celsius(current_temp_k)}{DEG_C} - {current_condition}"
    )
    expected_string_f = (
        f"{celsius_to_fahrenheit(kelvin_to_celsius(current_temp_k))}"
        f"{DEG_F} - {current_condition}"
    )
    assert weather_obj.get_current_weather() == expected_string_c
    assert weather_obj.get_current_weather(ScaleType.FAHRENHEIT) == expected_string_f


@pytest.mark.parametrize(
    "day, min_k, max_k",
    [
        (0, 284.4, 294.12),
        (1, 287.11, 293.79),
        (2, 284.23, 288.42),
    ],
)
def test_can_successfully_retrieve_temp_range(
    day: int, min_k: float, max_k: float, _setup_weather_fake_data: OpenWeatherMap
) -> None:
    """Test for retrieving temp ranges

    Args:
        day (int): Day of the week
        min_k (float): Minimum temperature in Kelvin
        max_k (float): Maximum temperature in Kelvin
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    min_c = str(kelvin_to_celsius(min_k)) + DEG_C
    max_c = str(kelvin_to_celsius(max_k)) + DEG_C
    min_f = str(celsius_to_fahrenheit(kelvin_to_celsius(min_k))) + DEG_F
    max_f = str(celsius_to_fahrenheit(kelvin_to_celsius(max_k))) + DEG_F
    assert weather_obj.get_temp_range(day) == f"{min_c} – {max_c}"
    assert weather_obj.get_temp_range(day, ScaleType.FAHRENHEIT) == f"{min_f} – {max_f}"


@pytest.mark.parametrize(
    "day, temp_k",
    [
        (0, 292.73),
        (1, 292.95),
        (2, 286.9),
    ],
)
def test_can_successfully_retrieve_future_weather(
    day: int, temp_k: float, _setup_weather_fake_data: OpenWeatherMap
) -> None:
    """Test for retrieving future weather

    Args:
        day (int): Day of the week
        temp_k (float): Temperature in Kelvin
        _setup_weather_fake_data (OpenWeatherMap): Fixture for weather data
    """
    weather_obj = _setup_weather_fake_data
    temp_c = str(kelvin_to_celsius(temp_k)) + DEG_C
    temp_f = str(celsius_to_fahrenheit(kelvin_to_celsius(temp_k))) + DEG_F
    assert weather_obj.get_future_weather(day) == temp_c
    assert weather_obj.get_future_weather(day, ScaleType.FAHRENHEIT) == temp_f



"""Desktop Display Driver

This class is a drop-in replacement for the Inky library InkyWHAT class.
Images are drawn using Pillow to a desktop window for testing purposes.
"""

from typing import Any, Dict, Optional

from PIL import Image, ImageDraw  # type: ignore


class DesktopDisplayDriver:
    """Desktop Display Driver."""

    WIDTH = 400
    HEIGHT = 300
    BORDER_SIZE = 5
    MAX_VAL = 255
    WHITE = (MAX_VAL, MAX_VAL, MAX_VAL, MAX_VAL)
    BLACK = (0, 0, 0, MAX_VAL)
    RED = (MAX_VAL, 0, 0, MAX_VAL)
    YELLOW = (MAX_VAL, MAX_VAL, MAX_VAL // 3, MAX_VAL)

    def __init__(self, base_color: str = "") -> None:
        """Initialize display driver.

        Args:
            base_color: base color
        """
        color_matcher: Dict[str, tuple] = {
            "black": self.BLACK,
            "white": self.WHITE,
            "red": self.RED,
            "yellow": self.YELLOW,
        }

        self._black: tuple = self.BLACK
        self._white: tuple = self.WHITE
        self._color: tuple = (
            self.BLACK if base_color == "" else color_matcher[base_color]
        )
        self._img: Optional[Image] = None
        self._img_draw: Optional[ImageDraw] = None

    def set_image(self, image: Any) -> None:
        """Set image

        Args:
            PIL image: image
        """
        self._img = image

    def set_border(self, border_color: Any) -> None:
        """Draw border around the edge (1/2 self.BORDER_SIZE + 1)

        Args:
            border_color: border color (one of self.<COLOR>)
        """
        side_width = self.BORDER_SIZE * 2
        left_width = side_width - 1
        bottom_width = side_width + 1

        if self._img:
            self._img_draw = ImageDraw.Draw(self._img)
            self._img_draw.line(
                (0, 0, 0, self.HEIGHT),  # |
                border_color,
                left_width,
            )
            self._img_draw.line(
                (0, self.HEIGHT, self.WIDTH, self.HEIGHT),  # |_
                border_color,
                bottom_width,
            )
            self._img_draw.line(
                (self.WIDTH, self.HEIGHT, self.WIDTH, 0),  # |_|
                border_color,
                side_width,
            )
            self._img_draw.line(
                (self.WIDTH, 0, 0, 0),  # □
                border_color,
                side_width,
            )

    def show(self) -> None:
        """Show image

        Uses Pillow to display image on desktop.
        """
        if self._img:
            self._img.show()
        else:
            raise RuntimeError("No image to show")
